'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var produce = require('immer-peasy');
var produce__default = _interopDefault(produce);
var React = require('react');
var React__default = _interopDefault(React);
var shallowEqual = _interopDefault(require('shallowequal'));
var redux = require('redux');
var reduxThunk = _interopDefault(require('redux-thunk'));
var memoizerific = _interopDefault(require('memoizerific'));

var StoreContext = React.createContext();

var isStateObject = function isStateObject(x) {
  return x !== null && typeof x === 'object' && !Array.isArray(x);
};
var get = function get(path, target) {
  return path.reduce(function (acc, cur) {
    return isStateObject(acc) ? acc[cur] : undefined;
  }, target);
};
var set = function set(path, target, value) {
  path.reduce(function (acc, cur, idx) {
    if (idx + 1 === path.length) {
      acc[cur] = value;
    } else {
      acc[cur] = acc[cur] || {};
    }

    return acc[cur];
  }, target);
};

function createStoreStateHook(Context) {
  return function useStoreState(mapState, dependencies) {
    if (dependencies === void 0) {
      dependencies = [];
    }

    var store = React.useContext(Context);

    var _useState = React.useState(function () {
      return mapState(store.getState());
    }),
        state = _useState[0],
        setState = _useState[1];

    var _useState2 = React.useState(null),
        error = _useState2[0],
        setError = _useState2[1]; // As our effect only fires on mount and unmount it won't have the state
    // changes visible to it, therefore we use a mutable ref to track this.


    var stateRef = React.useRef(state); // Helps avoid firing of events when unsubscribed, i.e. unmounted

    var isActive = React.useRef(true); // Tracks when a hooked component is unmounting

    var unmounted = React.useRef(false); // Throwing the error inline allows error boundaries the opportunity to
    // catch the error

    if (error) {
      throw error;
    }

    React.useEffect(function () {
      isActive.current = true;

      var calculateState = function calculateState() {
        if (!isActive.current) {
          return;
        }

        try {
          var newState = mapState(store.getState());

          if (newState === stateRef.current || isStateObject(newState) && isStateObject(stateRef.current) && shallowEqual(newState, stateRef.current)) {
            // Do nothing
            return;
          }

          stateRef.current = newState;
          setState(function () {
            return stateRef.current;
          });
        } catch (err) {
          // see https://github.com/reduxjs/react-redux/issues/1179
          // There is a possibility mapState will fail as the props/state that
          // the component has received is stale. Therefore we will afford the
          // application a small window of opportunity, where if they unmount
          // the component or provide it with new "valid" props (which will
          // subsequently cause a refreshed subscription cycle) then we will not
          // throw an error.
          // This is by no means a robust solution. We should track the
          // associated issue in the hope for a more dependable solution.
          // Setting the listener as "inactive", this can only be changed if the
          // incoming dependencies are different (i.e. props have changed)
          isActive.current = false;
          setTimeout(function () {
            if (!unmounted.current && !isActive.current) {
              setError(err);
            }
          }, 200); // give a window of opportunity
        }
      };

      calculateState();
      var unsubscribe = store.subscribe(calculateState);
      return function () {
        isActive.current = false;
        unsubscribe();
      };
    }, dependencies); // This effect will set the ref value to indicate that the component has
    // unmounted

    React.useEffect(function () {
      return function () {
        unmounted.current = true;
      };
    }, []);
    return state;
  };
}
var useStoreState = createStoreStateHook(StoreContext);
function createStoreActionsHook(Context) {
  return function useStoreActions(mapActions) {
    var store = React.useContext(Context);
    return mapActions(store.getActions());
  };
}
var useStoreActions = createStoreActionsHook(StoreContext);
function createStoreDispatchHook(Context) {
  return function useStoreDispatch() {
    var store = React.useContext(Context);
    return store.dispatch;
  };
}
var useStoreDispatch = createStoreDispatchHook(StoreContext);

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

var actionNameSymbol = '🙈actionName🙈';
var actionSymbol = '🙈action🙈';
var actionStateSymbol = '🙈actionState🙈';
var computedSymbol = '🙈computedSymbol🙈';
var computedConfigSymbol = '🙈computedConfigSymbol🙈';
var metaSymbol = '🙈meta🙈';
var thunkSymbol = '🙈thunk🙈';
var thunkStateSymbol = '🙈thunkState🙈';
var reducerSymbol = '🙈reducer🙈';

var actionName = function actionName(action) {
  return action[actionNameSymbol];
};
var debug = function debug(state) {
  if (produce.isDraft(state)) {
    var final = produce.finishDraft(produce.createDraft(state));
    return final;
  }

  return state;
};
var memo = function memo(fn, cacheSize) {
  return memoizerific(cacheSize)(fn);
};
var thunkStartName = function thunkStartName(action) {
  return action[actionNameSymbol] + "(started)";
};
var thunkCompleteName = function thunkCompleteName(action) {
  return action[actionNameSymbol] + "(completed)";
};
var thunkFailName = function thunkFailName(action) {
  return action[actionNameSymbol] + "(failed)";
};
var action = function action(fn, config) {
  fn[actionSymbol] = true;
  fn[actionStateSymbol] = {
    config: config
  };
  return fn;
};
var defaultStateResolvers = [function (state) {
  return state;
}];
var computed = function computed(fn, stateResolvers) {
  if (stateResolvers === void 0) {
    stateResolvers = defaultStateResolvers;
  }

  fn[computedSymbol] = true;
  fn[computedConfigSymbol] = {
    stateResolvers: stateResolvers
  };
  return fn;
};
var thunk = function thunk(fn, config) {
  fn[thunkSymbol] = true;
  fn[thunkStateSymbol] = {
    config: config
  };
  return fn;
};
var reducer = function reducer(fn) {
  fn[reducerSymbol] = true;
  return fn;
};

var tick = function tick() {
  return new Promise(function (resolve) {
    return setTimeout(resolve);
  });
};

var errorToPayload = function errorToPayload(err) {
  if (err instanceof Error) {
    return {
      message: err.message,
      stack: err.stack
    };
  }

  if (typeof err === 'string') {
    return err;
  }

  return undefined;
};

function createStoreInternals(_ref) {
  var initialState = _ref.initialState,
      injections = _ref.injections,
      model = _ref.model,
      reducerEnhancer = _ref.reducerEnhancer,
      references = _ref.references;
  var isInReducer = false;
  var defaultState = initialState || {};
  var actionCreatorDict = {};
  var actionCreators = {};
  var actionReducersDict = {};
  var actionThunks = {};
  var computedProperties = [];
  var customReducers = [];
  var listenerActionDefinitions = [];
  var listenerActionMap = {};

  var recursiveExtractDefsFromModel = function recursiveExtractDefsFromModel(current, parentPath) {
    return Object.keys(current).forEach(function (key) {
      var value = current[key];
      var path = [].concat(parentPath, [key]);
      var meta = {
        parent: parentPath,
        path: path
      };

      if (typeof value === 'function') {
        if (value[actionSymbol]) {
          var name = "@action." + path.join('.');
          value[actionNameSymbol] = name;
          value[metaSymbol] = meta; // Action Reducer

          var actionReducer = value;
          actionReducer[actionNameSymbol] = name;
          actionReducersDict[name] = actionReducer;

          var actionCreator = function actionCreator(payload) {
            var result = references.dispatch({
              type: actionReducer[actionNameSymbol],
              payload: payload
            });
            return result;
          };

          actionCreator[actionNameSymbol] = name;
          actionCreatorDict[name] = actionCreator;
          set(path, actionCreators, actionCreator);
          var config = value[actionStateSymbol].config;

          if (config && config.listenTo) {
            listenerActionDefinitions.push(value);
          }
        } else if (value[thunkSymbol]) {
          var _name = "@thunk." + path.join('.');

          value[actionNameSymbol] = _name; // Thunk Action

          var action = function action(payload) {
            return value(get(parentPath, actionCreators), payload, {
              dispatch: references.dispatch,
              getState: function getState() {
                return get(parentPath, references.getState());
              },
              getStoreActions: function getStoreActions() {
                return actionCreators;
              },
              getStoreState: references.getState,
              injections: injections,
              meta: meta
            });
          };

          set(path, actionThunks, action); // Thunk Action Creator

          var _actionCreator = function _actionCreator(payload) {
            return tick().then(function () {
              return references.dispatch({
                type: _name + "(started)",
                payload: payload
              });
            }).then(function () {
              return references.dispatch(function () {
                return action(payload);
              });
            }).then(function (result) {
              references.dispatch({
                type: _name + "(completed)",
                payload: payload
              });
              return result;
            }).catch(function (err) {
              references.dispatch({
                type: _name + "(failed)",
                payload: payload,
                error: errorToPayload(err)
              });
              throw err;
            });
          };

          _actionCreator[actionNameSymbol] = _name;
          actionCreatorDict[_name] = _actionCreator;
          set(path, actionCreators, _actionCreator);
          var _config = value[thunkStateSymbol].config;

          if (_config && _config.listenTo) {
            listenerActionDefinitions.push(value);
          }
        } else if (value[computedSymbol]) {
          var target = get(parentPath, defaultState);

          if (!target) {
            target = {};
            set(parentPath, defaultState, target);
          }

          var _config2 = value[computedConfigSymbol];
          var stateResolvers = _config2.stateResolvers;
          var memoisedResultFn = memoizerific(1)(value);
          var cache;

          var createComputedProperty = function createComputedProperty(o) {
            Object.defineProperty(o, key, {
              configurable: true,
              enumerable: true,
              get: function get$1() {
                var storeState = isInReducer ? references.currentState : references.getState();

                var state = get(parentPath, storeState);

                var inputs = stateResolvers.map(function (resolver) {
                  return resolver(state, storeState);
                });
                cache = memoisedResultFn.apply(void 0, inputs);
                return cache;
              },
              set: function set() {
                throw new Error("Easy Peasy: You attempted to set \"" + path.join('.') + "\", which is a computed property set a computed property");
              }
            });
          };

          createComputedProperty(target);
          computedProperties.push({
            key: key,
            parentPath: parentPath,
            createComputedProperty: createComputedProperty
          });
        } else if (value[reducerSymbol]) {
          customReducers.push({
            path: path,
            reducer: value
          });
        }
      } else if (isStateObject(value) && Object.keys(value).length > 0) {
        var existing = get(path, defaultState);

        if (existing == null) {
          set(path, defaultState, {});
        }

        recursiveExtractDefsFromModel(value, path);
      } else {
        // State
        var initialParentRef = get(parentPath, initialState);

        if (initialParentRef && key in initialParentRef) {
          set(path, defaultState, initialParentRef[key]);
        } else {
          set(path, defaultState, value);
        }
      }
    });
  };

  recursiveExtractDefsFromModel(model, []);
  listenerActionDefinitions.forEach(function (listenerAction) {
    var _ref2 = listenerAction[actionStateSymbol] || listenerAction[thunkStateSymbol],
        listenTo = _ref2.config.listenTo;

    var targetName;

    var processListenTo = function processListenTo(target) {
      if (typeof target === 'function' && target[actionNameSymbol] && actionCreatorDict[target[actionNameSymbol]]) {
        if (target[thunkSymbol]) {
          targetName = thunkCompleteName(target);
        } else {
          targetName = target[actionNameSymbol];
        }
      } else if (typeof target === 'string') {
        targetName = target;
      }

      var listenerReg = listenerActionMap[targetName] || [];
      listenerReg.push(actionCreatorDict[listenerAction[actionNameSymbol]]);
      listenerActionMap[targetName] = listenerReg;
    };

    if (Array.isArray(listenTo)) {
      listenTo.forEach(processListenTo);
    } else {
      processListenTo(listenTo);
    }
  });

  var createReducer = function createReducer() {
    var runActionReducerAtPath = function runActionReducerAtPath(state, action, actionReducer, path) {
      var current = get(path, state);

      return path.length === 0 ? produce__default(state, function (_draft) {
        return actionReducer(_draft, action.payload);
      }) : produce__default(state, function (draft) {
        set(actionReducer[metaSymbol].parent, draft, produce__default(current, function (_draft) {
          return actionReducer(_draft, action.payload);
        }));
      });
    };

    var reducerForActions = function reducerForActions(state, action) {
      var actionReducer = actionReducersDict[action.type];

      if (actionReducer) {
        return runActionReducerAtPath(state, action, actionReducer, actionReducer[metaSymbol].parent);
      }

      return state;
    };

    var reducerForCustomReducers = function reducerForCustomReducers(state, action) {
      return produce__default(state, function (draft) {
        customReducers.forEach(function (_ref3) {
          var p = _ref3.path,
              red = _ref3.reducer;

          var current = get(p, draft);

          set(p, draft, red(current, action));
        });
      });
    };

    var rootReducer = function rootReducer(state, action) {
      isInReducer = true;
      var stateAfterActions = reducerForActions(state, action);
      var result = reducerForCustomReducers(stateAfterActions, action);
      isInReducer = false;

      if (result !== state) {
        computedProperties.forEach(function (_ref4) {
          var parentPath = _ref4.parentPath,
              createComputedProperty = _ref4.createComputedProperty;
          createComputedProperty(get(parentPath, result));
        });
      }

      return result;
    };

    return rootReducer;
  };

  return {
    actionCreators: actionCreators,
    defaultState: defaultState,
    listenerActionMap: listenerActionMap,
    reducer: reducerEnhancer(createReducer())
  };
}

function createStore(model, options) {
  if (options === void 0) {
    options = {};
  }

  var _options = options,
      compose = _options.compose,
      _options$devTools = _options.devTools,
      devTools = _options$devTools === void 0 ? true : _options$devTools,
      _options$enhancers = _options.enhancers,
      enhancers = _options$enhancers === void 0 ? [] : _options$enhancers,
      _options$initialState = _options.initialState,
      initialState = _options$initialState === void 0 ? {} : _options$initialState,
      injections = _options.injections,
      _options$middleware = _options.middleware,
      middleware = _options$middleware === void 0 ? [] : _options$middleware,
      _options$mockActions = _options.mockActions,
      mockActions = _options$mockActions === void 0 ? false : _options$mockActions,
      _options$name = _options.name,
      storeName = _options$name === void 0 ? "EasyPeasyStore" : _options$name,
      _options$reducerEnhan = _options.reducerEnhancer,
      reducerEnhancer = _options$reducerEnhan === void 0 ? function (rootReducer) {
    return rootReducer;
  } : _options$reducerEnhan;

  var modelDefinition = _extends({}, model, {
    logFullState: thunk(function (actions, payload, _ref) {
      var getState = _ref.getState;
      // eslint-disable-next-line no-console
      console.log(JSON.stringify(getState(), null, 2));
    }),
    replaceState: action(function (state, payload) {
      return payload;
    })
  });

  var references = {};
  var mockedActions = [];
  var composeEnhancers = compose || (devTools && typeof window !== 'undefined' && window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ ? window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__({
    name: storeName
  }) : redux.compose);

  var bindStoreInternals = function bindStoreInternals(state) {
    references.internals = createStoreInternals({
      initialState: state,
      injections: injections,
      model: modelDefinition,
      reducerEnhancer: reducerEnhancer,
      references: references
    });
  };

  bindStoreInternals(initialState);

  var listenerActionsMiddleware = function listenerActionsMiddleware() {
    return function (next) {
      return function (action) {
        var result = next(action);

        if (action && references.internals.listenerActionMap[action.type] && references.internals.listenerActionMap[action.type].length > 0) {
          references.internals.listenerActionMap[action.type].forEach(function (actionCreator) {
            actionCreator(action.payload);
          });
        }

        return result;
      };
    };
  };

  var mockActionsMiddleware = function mockActionsMiddleware() {
    return function (next) {
      return function (action) {
        if (mockActions) {
          if (action == null || typeof action === 'object' && action.type === '@@EP/LISTENER') ; else {
            mockedActions.push(action);
          }

          return undefined;
        }

        return next(action);
      };
    };
  };

  var currentStateMiddleware = function currentStateMiddleware() {
    return function (next) {
      return function (action) {
        references.currentState = references.getState();
        return next(action);
      };
    };
  };

  var store = redux.createStore(references.internals.reducer, references.internals.defaultState, composeEnhancers.apply(void 0, [redux.applyMiddleware.apply(void 0, [reduxThunk].concat(middleware, [listenerActionsMiddleware, currentStateMiddleware, mockActionsMiddleware]))].concat(enhancers)));

  store.getMockedActions = function () {
    return [].concat(mockedActions);
  };

  store.clearMockedActions = function () {
    mockedActions = [];
  };

  store.getActions = function () {
    return references.internals.actionCreators;
  };

  references.dispatch = store.dispatch;
  references.getState = store.getState;

  var rebindStore = function rebindStore(removeKey) {
    var currentState = store.getState();

    if (removeKey) {
      delete currentState[removeKey];
    }

    bindStoreInternals(store.getState());
    store.replaceReducer(references.internals.reducer);
    store.getActions().replaceState(references.internals.defaultState);
  };

  store.addModel = function (key, modelForKey) {
    if (modelDefinition[key] && process.env.NODE_ENV !== 'production') {
      // eslint-disable-next-line no-console
      console.warn("easy-peasy: The store model already contains a model definition for \"" + key + "\"");
      store.removeModel(key);
    }

    modelDefinition[key] = modelForKey;
    rebindStore();
  };

  store.removeModel = function (key) {
    if (!modelDefinition[key]) {
      if (process.env.NODE_ENV !== 'production') {
        // eslint-disable-next-line no-console
        console.warn("easy-peasy: The store model does not contain a model definition for \"" + key + "\"");
      }

      return;
    }

    delete modelDefinition[key];
    rebindStore(key);
  };

  store.useStoreActions = useStoreActions;
  store.useStoreDispatch = useStoreDispatch;
  store.useStoreState = useStoreState;
  return store;
}

/* eslint-disable react/prop-types */
function createContextStore(model, config) {
  var StoreContext = React.createContext();

  function Provider(_ref) {
    var children = _ref.children,
        initialData = _ref.initialData;
    var store = React.useMemo(function () {
      return createStore(typeof model === 'function' ? model(initialData) : model, config);
    }, []);
    return React__default.createElement(StoreContext.Provider, {
      value: store
    }, children);
  }

  function useStore() {
    return React.useContext(StoreContext);
  }

  return {
    Provider: Provider,
    useStore: useStore,
    useStoreState: createStoreStateHook(StoreContext),
    useStoreActions: createStoreActionsHook(StoreContext),
    useStoreDispatch: createStoreDispatchHook(StoreContext)
  };
}

/**
 * Some good references on the topic of reinitialisation:
 * - https://github.com/facebook/react/issues/14830
 */

function createComponentStore(model, config) {
  return function useLocalStore(initialData) {
    var store = React.useMemo(function () {
      return createStore(typeof model === 'function' ? model(initialData) : model, config);
    }, []);
    var previousStateRef = React.useRef(store.getState());

    var _useState = React.useState(function () {
      return store.getState();
    }),
        currentState = _useState[0],
        setCurrentState = _useState[1];

    React.useEffect(function () {
      return store.subscribe(function () {
        var nextState = store.getState();

        if (previousStateRef.current !== nextState) {
          previousStateRef.current = nextState;
          setCurrentState(nextState);
        }
      });
    }, [store]);
    return [currentState, store.getActions()];
  };
}

var StoreProvider = function StoreProvider(_ref) {
  var children = _ref.children,
      store = _ref.store;
  return React__default.createElement(StoreContext.Provider, {
    value: store
  }, children);
};

/**
 * The auto freeze feature of immer doesn't seem to work in our testing. We have
 * explicitly disabled it to avoid perf issues.
 */

produce.setAutoFreeze(false);

exports.StoreProvider = StoreProvider;
exports.action = action;
exports.actionName = actionName;
exports.computed = computed;
exports.createComponentStore = createComponentStore;
exports.createContextStore = createContextStore;
exports.createStore = createStore;
exports.debug = debug;
exports.memo = memo;
exports.reducer = reducer;
exports.thunk = thunk;
exports.thunkCompleteName = thunkCompleteName;
exports.thunkFailName = thunkFailName;
exports.thunkStartName = thunkStartName;
exports.useStoreActions = useStoreActions;
exports.useStoreDispatch = useStoreDispatch;
exports.useStoreState = useStoreState;
//# sourceMappingURL=easy-peasy.cjs.js.map
